/* USB EHCI Host Mass Storaage Testing for Teensy 3.6
 * Copyright 2017 Warren Watson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <Arduino.h>
#include "USBHost_t36.h"
#include "MassStorage.h"

#define print   USBHost::print_
#define println USBHost::println_

// Big Endian/Little Endian
#define swap32(x) ((x >> 24) & 0xff) | \
				  ((x << 8) & 0xff0000) | \
				  ((x >> 8) & 0xff00) |  \
                  ((x << 24) & 0xff000000)


void msController::init()
{
	contribute_Pipes(mypipes, sizeof(mypipes)/sizeof(Pipe_t));
	contribute_Transfers(mytransfers, sizeof(mytransfers)/sizeof(Transfer_t));
	driver_ready_for_device(this);
}

bool msController::claim(Device_t *dev, int type, const uint8_t *descriptors, uint32_t len)
{
	println("msController claim this=", (uint32_t)this, HEX);
	// only claim at interface level

	if (type != 1) return false;
	if (len < 9+7+7) return false; // Interface descriptor + 2 endpoint decriptors 

	print_hexbytes(descriptors, len);

	uint32_t numendpoint = descriptors[4];
	if (numendpoint < 1) return false; 
	if (descriptors[5] != 8) return false; // bInterfaceClass, 8 = MASS Storage class
	if (descriptors[6] != 6) return false; // bInterfaceSubClass, 6 = SCSI transparent command set (SCSI Standards)
	if (descriptors[7] != 80) return false; // bInterfaceProtocol, 80 = BULK-ONLY TRANSPORT
	if (descriptors[10] != 5 || descriptors[17] != 5) return false; // Must have bulk-in and bulk-out endpoints
	uint8_t endpointType = (descriptors[12] | descriptors[19]) & 0x03; 
	endpointIn = descriptors[11]; // bulk-in descriptor 1 81h
	endpointOut = descriptors[18]; // bulk-out descriptor 2 02h
	if ((endpointIn & 0xF0) != 0x80) return false; // must be IN direction
	if ((endpointOut & 0xF0) != 0x00) return false; // must be OUT direction
	uint32_t sizeIn = descriptors[13] | (descriptors[14] << 8);
	uint32_t sizeOut = descriptors[20] | (descriptors[21] << 8);
	packetSizeIn = sizeIn;	
	packetSizeOut = sizeOut;	
	uint32_t intervalIn = descriptors[15];
	uint32_t intervalOut = descriptors[22];
	datapipeIn = new_Pipe(dev, 2, endpointIn, 1, packetSizeIn, intervalIn);
	datapipeOut = new_Pipe(dev, 2, endpointOut, 0, packetSizeOut, intervalOut);
	datapipeIn->callback_function = callbackIn;
	datapipeOut->callback_function = callbackOut;
	msOutCompleted = false;
	msInCompleted = false;
	deviceAvailable = true;
	return true;
}

void msController::control(const Transfer_t *transfer)
{
	println("control Callback (static)");
	if (transfer->driver) {
		((msController *)(transfer->driver))->new_data(transfer);
	}
}

void msController::callbackIn(const Transfer_t *transfer)
{
	println("msController CallbackIn (static)");
	if (transfer->driver) {
		((msController *)(transfer->driver))->new_dataIn(transfer);
	}
}

void msController::callbackOut(const Transfer_t *transfer)
{
	println("msController CallbackOut (static)");
	if (transfer->driver) {
		((msController *)(transfer->driver))->new_dataOut(transfer);
	}
}

void msController::disconnect()
{
	deviceAvailable = false;
	println("Device Disconnected...");
	// TODO: free resources
}

void msController::new_data(const Transfer_t *transfer)
{
	println("msController Control Callback (member)");
	print("  msData: ");
	print_hexbytes(transfer->buffer, 8);
	memcpy(report,transfer->buffer,8);
}

void msController::new_dataIn(const Transfer_t *transfer)
{
	msInCompleted = true;
}

void msController::new_dataOut(const Transfer_t *transfer)
{
	msOutCompleted = true;
}

//---------------------------------------------------------------------------
// Perform Mass Storage Reset
//---------------------------------------------------------------------------
void msController::msReset() {
	mk_setup(setup, 0x21, 0xff, 0, 0, 0);
	queue_Control_Transfer(device, &setup, NULL, this);
}

//---------------------------------------------------------------------------
// Get MAX LUN (flakey results)
//---------------------------------------------------------------------------
uint8_t msController::msGetMaxLun() {
	mk_setup(setup, 0xa1, 0xfe, 0, 0, 1);
	queue_Control_Transfer(device, &setup, report, this);
	delay(1); 
	return report[0];
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void msController::WaitMediaReady() {
	uint8_t msResult = 0;
	msResult = msStartStopUnit(1); // Spin up drive if stopped
	do {
		msResult = msTestReady();
	} while(msResult);
// TODO: Process Timeout
}

//---------------------------------------------------------------------------
// Execute SCSI Command
//---------------------------------------------------------------------------
uint8_t msController::msDoCommand(msCommandBlockWrapper_t *CBW,	void *buffer)
{
	if(CBWTag == 0xFFFFFFFF)
		CBWTag = 1;
	if((CBW->Flags == CMDDIRDATAIN)) {
		queue_Data_Transfer(datapipeIn, buffer, CBW->TransferLength, this);
		queue_Data_Transfer(datapipeOut, CBW, sizeof(msCommandBlockWrapper_t), this);
	} else {
		queue_Data_Transfer(datapipeOut, CBW, sizeof(msCommandBlockWrapper_t), this);
		queue_Data_Transfer(datapipeOut, buffer, CBW->TransferLength, this);
	}
	while(!msOutCompleted) {
		delay(0); // Why is this delay needed??
	}
	msOutCompleted = false;
	return msGetCSW();
}

//---------------------------------------------------------------------------
// Get Command Status Wrapper
//---------------------------------------------------------------------------
uint8_t msController::msGetCSW() {
	uint8_t CSWResult = 0;
	msCommandStatusWrapper_t CSW = (msCommandStatusWrapper_t)
	{
		.Signature = CSWSIGNATURE,
		.Tag = 0,
		.DataResidue = 0,
		.Status = 0
	};
	queue_Data_Transfer(datapipeIn, &CSW, sizeof(CSW), this);
	while(!msInCompleted){
		delay(0); // Why is this delay needed??
//TODO: Do a timeout check.
	}
	msInCompleted = false;
	print("CommandStatusWrapper: ");
	print_hexbytes(&CSW, sizeof(CSW));
	CSWResult = checkCSW(&CSW);
	if(CSWResult) {
		return CSWResult;
	}
	return CSW.Status;
}

//---------------------------------------------------------------------------
// Check for valid CSW
// Sometimes an incomplete CSW is returned. A valid CSW signature is returned
// and zero is returned in the CSW tag field and the status field is zero.
// Not sure why. Short read? 
//---------------------------------------------------------------------------
uint8_t msController::checkCSW(msCommandStatusWrapper_t *CSW) {

	if(CSW->Signature != CSWSIGNATURE) {
		println("CSW:Sig error\r\n");
		return MSCSWSIGERROR;
	}
	if(CSW->Tag != CBWTag) {
		println("CSW:Wrong tag\r\n");
		return MSCSWTAGERROR;
	}
	return MSCOMMANDPASS;
}

//---------------------------------------------------------------------------
// Test Unit Ready
//---------------------------------------------------------------------------
uint8_t msController::msTestReady() {
	msCommandBlockWrapper_t CBW = (msCommandBlockWrapper_t)
	{
		.Signature          = CBWSIGNATURE,
		.Tag                = ++CBWTag,
		.TransferLength     = 0,
		.Flags              = CMDDIRDATAIN,
		.LUN                = 0,
		.CommandLength      = 6,
		.CommandData        = {CMDTESTUNITREADY, 0x00, 0x00, 0x00, 0x00, 0x00}
	};
	queue_Data_Transfer(datapipeOut, &CBW, sizeof(CBW), this);
	return msGetCSW();
}

//---------------------------------------------------------------------------
// Start/Stop unit
//---------------------------------------------------------------------------
uint8_t msController::msStartStopUnit(uint8_t mode) {
	msCommandBlockWrapper_t CBW = (msCommandBlockWrapper_t)
	{
		.Signature          = CBWSIGNATURE,
		.Tag                = ++CBWTag,
		.TransferLength     = 0,
		.Flags              = CMDDIRDATAIN,
		.LUN                = 0,
		.CommandLength      = 6,
		.CommandData        = {CMDSTARTSTOPUNIT, 0x00, 0x00, 0x00, mode, 0x00}
	};
	queue_Data_Transfer(datapipeOut, &CBW, sizeof(CBW), this);
	return msGetCSW();
}

//---------------------------------------------------------------------------
// Read Mass Storage Device Capacity (Number of Blocks and Block Size)
//---------------------------------------------------------------------------
uint8_t msController::msReadDeviceCapacity(msSCSICapacity_t * const Capacity)
{
	uint8_t result = 0;
//	WaitMediaReady();
	msCommandBlockWrapper_t CBW = (msCommandBlockWrapper_t)
	{
		.Signature          = CBWSIGNATURE,
		.Tag                = ++CBWTag,
		.TransferLength     = sizeof(msSCSICapacity_t),
		.Flags              = CMDDIRDATAIN,
		.LUN                = 0,
		.CommandLength      = 10,
		.CommandData        = {CMDRDCAPACITY10,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}
	};
	result = msDoCommand(&CBW, Capacity);
	Capacity->Blocks = swap32(Capacity->Blocks);
	Capacity->BlockSize = swap32(Capacity->BlockSize);
	return result;
}

//---------------------------------------------------------------------------
// Do Mass Storage Device Inquiry
//---------------------------------------------------------------------------
uint8_t msController::msDeviceInquiry(msInquiryResponse_t * const Inquiry)
{
//	WaitMediaReady();
	msCommandBlockWrapper_t CBW = (msCommandBlockWrapper_t)
	{
		.Signature          = CBWSIGNATURE,
		.Tag                = ++CBWTag,
		.TransferLength     = sizeof(msInquiryResponse_t),
		.Flags              = CMDDIRDATAIN,
		.LUN                = 0,
		.CommandLength      = 6,
		.CommandData        = {CMDINQUIRY,0x00,0x00,0x00,0x00,0x00}
	};
	return msDoCommand(&CBW, Inquiry);
}

//---------------------------------------------------------------------------
// Request Sense Data
//---------------------------------------------------------------------------
uint8_t msController::msRequestSense(msRequestSenseResponse_t * const Sense)
{
//	WaitMediaReady();
	msCommandBlockWrapper_t CBW = (msCommandBlockWrapper_t)
	{
		.Signature          = CBWSIGNATURE,
		.Tag                = ++CBWTag,
		.TransferLength     = sizeof(msRequestSenseResponse_t),
		.Flags              = CMDDIRDATAIN,
		.LUN                = 0,
		.CommandLength      = 6,
		.CommandData        = {CMDREQUESTSENSE, 0x00, 0x00, 0x00, sizeof(msRequestSenseResponse_t), 0x00}
	};
	return msDoCommand(&CBW, Sense);
}

//---------------------------------------------------------------------------
// Report LUNs
//---------------------------------------------------------------------------
uint8_t msController::msReportLUNs(uint8_t *Buffer)
{
//	WaitMediaReady();
	msCommandBlockWrapper_t CBW = (msCommandBlockWrapper_t)
	{
		.Signature          = CBWSIGNATURE,
		.Tag                = ++CBWTag,
		.TransferLength     = 512,
		.Flags              = CMDDIRDATAIN,
		.LUN                = 0,
		.CommandLength      = 8,
		.CommandData        = {CMDREPORTLUNS, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}
	};
	return msDoCommand(&CBW, Buffer);
}

//---------------------------------------------------------------------------
// Read Sectors
//---------------------------------------------------------------------------
uint8_t msController::msReadBlocks(const uint32_t BlockAddress,
	const uint8_t Blocks, const uint16_t BlockSize,	void * sectorBuffer) {
	msCommandBlockWrapper_t CBW = (msCommandBlockWrapper_t)
	{
		.Signature          = CBWSIGNATURE,
		.Tag                = ++CBWTag,
		.TransferLength     = (uint32_t)(Blocks * BlockSize),
		.Flags              = CMDDIRDATAIN,
		.LUN                = 0,
		.CommandLength      = 10,
		.CommandData        = {CMDRD10, 0x00,
							  (uint8_t)(BlockAddress >> 24),
							  (uint8_t)(BlockAddress >> 16),
							  (uint8_t)(BlockAddress >> 8),
							  (uint8_t)(BlockAddress & 0xFF),
							   0x00, 0x00, Blocks, 0x00}
	};
	return msDoCommand(&CBW, sectorBuffer);
}

//---------------------------------------------------------------------------
// Write Sectors
//---------------------------------------------------------------------------
uint8_t msController::msWriteBlocks(const uint32_t BlockAddress,
	const uint8_t Blocks, const uint16_t BlockSize, void * sectorBuffer) {
//	WaitMediaReady();
	msCommandBlockWrapper_t CBW = (msCommandBlockWrapper_t)
	{
		.Signature          = CBWSIGNATURE,
		.Tag                = ++CBWTag,
		.TransferLength     = (uint32_t)(Blocks * BlockSize),
		.Flags              = CMDDIRDATAOUT,
		.LUN                = 0,
		.CommandLength      = 10,
		.CommandData        = {CMDWR10, 0x00,
		                      (uint8_t)(BlockAddress >> 24),
							  (uint8_t)(BlockAddress >> 16),
							  (uint8_t)(BlockAddress >> 8),
							  (uint8_t)(BlockAddress & 0xFF),
							  0x00, 0x00, Blocks, 0x00}
	};
	return msDoCommand(&CBW, sectorBuffer);
}
