/* USB EHCI Host Mass Storaage Testing for Teensy 3.6
 * Copyright 2017 Warren Watson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef _MASS_STORAGE_H_
#define _MASS_STORAGE_H_

#include "USBHost_t36.h"

// SCSI Commands
#define CBWSIGNATURE	 0x43425355UL // 'USBC'
#define CSWSIGNATURE	 0x53425355UL // 'USBS'
#define CMDDIRDATAOUT	 0x00
#define CMDDIRDATAIN	 0x80
#define CMDRDCAPACITY10	 0x25
#define CMDINQUIRY       0x12
#define CMDRD10			 0x28
#define CMDWR10			 0x2A
#define CMDTESTUNITREADY 0x00
#define CMDREQUESTSENSE  0x03
#define CMDSTARTSTOPUNIT 0x1B
#define CMDREPORTLUNS    0xA0

// Command Status Wrapper Error Codes
#define	MSCOMMANDPASS 		0
#define	MSCOMMANDFAIL  		1
#define	MSCOMMANDPHASEERROR	2
#define MSCSWTAGERROR  		3
#define MSCSWSIGERROR		4

// Sense Key Codes
#define NO_SENSE            0x00
#define RECOVERED_ERROR     0x01
#define NOT_READY           0x02
#define MEDIUM_ERROR        0x03
#define HARDWARE_ERROR      0x04
#define ILLEGAL_REQUEST     0x05
#define UNIT_ATTENTION      0x06
#define DATA_PROTECT        0x07
#define BLANK_CHECK         0x08
#define VENDOR_SPECIFIC     0x09
#define COPY_ABORTED        0x0A
#define ABORTED_COMMAND     0x0B
#define VOLUME_OVERFLOW     0x0D
#define MISCOMPARE          0x0E

// Command Block Wrapper Struct
typedef struct
{
	uint32_t Signature;
	uint32_t Tag;
	uint32_t TransferLength;
	uint8_t Flags;
	uint8_t LUN;
	uint8_t CommandLength;
	uint8_t CommandData[16];
}  __attribute__((packed)) msCommandBlockWrapper_t;

// Command Status Wrapper Struct
typedef struct
{
	uint32_t Signature;
	uint32_t Tag;
	uint32_t DataResidue;
	uint8_t  Status;
}  __attribute__((packed)) msCommandStatusWrapper_t;

// Device Capacity Struct
typedef struct
{
	uint32_t Blocks;
	uint32_t BlockSize;
} msSCSICapacity_t;

//Inquiry Command Reponse Struct
typedef struct
{
	unsigned DeviceType          : 5;
	unsigned PeripheralQualifier : 3;
	unsigned Reserved            : 7;
	unsigned Removable           : 1;
	uint8_t  Version;
	unsigned ResponseDataFormat  : 4;
	unsigned Reserved2           : 1;
	unsigned NormACA             : 1;
	unsigned TrmTsk              : 1;
	unsigned AERC                : 1;
	uint8_t  AdditionalLength;
	uint8_t  Reserved3[2];
	unsigned SoftReset           : 1;
	unsigned CmdQue              : 1;
	unsigned Reserved4           : 1;
	unsigned Linked              : 1;
	unsigned Sync                : 1;
	unsigned WideBus16Bit        : 1;
	unsigned WideBus32Bit        : 1;
	unsigned RelAddr             : 1;
	uint8_t  VendorID[8];
	uint8_t  ProductID[16];
	uint8_t  RevisionID[4];
}  __attribute__((packed)) msInquiryResponse_t;

// Request Sense Response Struct
typedef struct
{
	uint8_t  ResponseCode;

	uint8_t  SegmentNumber;

	unsigned SenseKey            : 4;
	unsigned Reserved            : 1;
	unsigned ILI                 : 1;
	unsigned EOM                 : 1;
	unsigned FileMark            : 1;

	uint8_t  Information[4];
	uint8_t  AdditionalLength;
	uint8_t  CmdSpecificInformation[4];
	uint8_t  AdditionalSenseCode;
	uint8_t  AdditionalSenseQualifier;
	uint8_t  FieldReplaceableUnitCode;
	uint8_t  SenseKeySpecific[3];
}  __attribute__((packed)) msRequestSenseResponse_t;

//--------------------------------------------------------------------------
class msController : public USBDriver {
public:
	msController(USBHost &host) { init(); }
	msController(USBHost *host) { init(); }
	void msReset();
	uint8_t msGetMaxLun();
	bool available() { delay(0); return deviceAvailable; }
	void WaitMediaReady();
	uint8_t msTestReady();
	uint8_t msReportLUNs(uint8_t *Buffer);
	uint8_t msStartStopUnit(uint8_t mode);
	uint8_t msReadDeviceCapacity(msSCSICapacity_t * const Capacity);
	uint8_t msDeviceInquiry(msInquiryResponse_t * const Inquiry);
	uint8_t msRequestSense(msRequestSenseResponse_t * const Sense);
	uint8_t msReadBlocks(const uint32_t BlockAddress, const uint8_t Blocks,
						 const uint16_t BlockSize, void * sectorBuffer);
	uint8_t msWriteBlocks(const uint32_t BlockAddress, const uint8_t Blocks,
                        const uint16_t BlockSize,	void * sectorBuffer);
protected:
	virtual bool claim(Device_t *device, int type, const uint8_t *descriptors, uint32_t len);
	virtual void control(const Transfer_t *transfer);
	virtual void disconnect();
	static void callbackIn(const Transfer_t *transfer);
	static void callbackOut(const Transfer_t *transfer);
	void new_data(const Transfer_t *transfer);
	void new_dataIn(const Transfer_t *transfer);
	void new_dataOut(const Transfer_t *transfer);
	void init();
	uint8_t msDoCommand(msCommandBlockWrapper_t *CBW, void *buffer);
	uint8_t msGetCSW();
	uint8_t checkCSW(msCommandStatusWrapper_t *CSW);
private:
	Pipe_t mypipes[3] __attribute__ ((aligned(32)));
	Transfer_t mytransfers[7] __attribute__ ((aligned(32)));
	strbuf_t mystring_bufs[1];
	uint32_t packetSizeIn;
	uint32_t packetSizeOut;
	Pipe_t *datapipeIn;
	Pipe_t *datapipeOut;
	setup_t setup;
	uint8_t maxLUN = 0;
	bool msOutCompleted = false;
	bool msInCompleted = false;
	uint32_t CBWTag = 0;
	uint32_t endpointIn = 0;
	uint32_t endpointOut = 0;
	uint8_t report[8];
	uint8_t reportIn[512];
	uint8_t reportOut[512];
	bool deviceAvailable = false;
};

#endif //_MASS_STORAGE_H_
